// Razred Drevo je predstavljen s podatkom
// kje je korensko vozlisce

public class Drevo{

   //referenca na korensko vozlisce
   private  Vozel koren;

   //konstruktor, ki ustvari prazno drevo
   public Drevo(){
      this.koren = null;
   }

   //konstruktor, ki ustavri drevo z vozlom v
   public Drevo(Vozel v){
      this.koren = v;
   }


   //metoda, ki ugotovi ali je drevo prazno
   public boolean prazno(){
      return this.koren == null;//vrne true ali false
   }
   /***********************************************************************/
   public Vozel koren(){
      return this.koren;//vrne koren drevesa
   }
   /**********************************************************************/

   //metoda vrne drevo
   public int vrni()throws Exception{
      if(!this.prazno()){ //ce ni prazno ga vrne
         return this.koren.vrniPodatek();
      }
      throw new Exception("Drevo je prazno! ");//drugace vrze izjemo
   }

   //metoda, ki vrne levega sina (levo poddrevo)
   public Drevo leviSin(){
      Drevo novo = new Drevo();// ustvarimo novo prazno drevo
      if(!this.prazno()){// ce ni prazno
         novo.koren = this.koren.vrniLevi();//koren levega postane koren novega
      }
      return novo;//drugace vrne novo prazno drevo
   }

   //metoda, ki vrne desnega sina (desno poddrevo)
   public Drevo desniSin(){
      Drevo novo = new Drevo();// ustvarimo novo prazno drevo
      if(!this.prazno()){// ce ni prazno
         novo.koren = this.koren.vrniDesni();//koren desnega postane koren novega
      }
      return novo;//drugace vrne novo prazno drevo
   }

   //metoda, ki vrne oeta
   public Drevo oce(){
      Drevo novo = new Drevo();//novo prazno drevo
      if(!this.prazno()){// ce ni prazno
         novo.koren = this.koren.vrniOce();//koren oceta postane koren novega
      }
      return novo;//drugace vrne novo prazno drevo
   }


//************************************************************************//

   // metoda, ki vrne stevilo vozlisc v dvojiskem drevesu
   public int stevilo_vozlisc(){
         if(prazno()){ // ce prazno
         return 0; //nima nobenega vozlisca
         }
         // vsota vozlisc v levem poddrevesu plus vsota vozlisc v desnem plus koren drevesa
         else return leviSin().stevilo_vozlisc()+ desniSin().stevilo_vozlisc()+ 1;
   }

   // metoda visina, ki doloci visino dvojiskega drevesa
   public int visina(){
     if(prazno()){ //ce prazno
        //visina praznega drevesa je nic
        return 0;
     }
     //maksimum od obeh visin, levega in desnega poddrevesa plus koren
     else return  Math.max(leviSin().visina(), desniSin().visina()) + 1;
   }

   // metoda, s katero poiscemo element v splosnem dvojiskem drevesu
   public  static boolean jeVDrevesu(Drevo drevo, int elt)throws Exception{
      if(drevo.prazno()) return false; // v praznem drevesu elementa ni
      if(drevo.vrni() == elt) return true; //element smo nasli
      return(jeVDrevesu(drevo.leviSin(), elt) || // iscemo levo
                 jeVDrevesu(drevo.desniSin(),elt)); // in desno
   }

   // metoda s katero poiscemo element v iskalnem dvojiskem drevesu
   public static boolean jeVIskalnemDrevesu(Drevo drevo, int elt)throws Exception{
      // ali v iskalnem drevesu drevo obstaja element elt
      if (drevo.prazno()) return false; // v praznem drevesu ni elementa
      if (drevo.vrni() == elt) return true; // element smo nasli
      if (elt < drevo.vrni() ) return jeVIskalnemDrevesu(drevo.leviSin(), elt);
          // iscemo levo
      return jeVIskalnemDrevesu(drevo.desniSin(),elt);      // iscemo desno
   }

   // metoda, ki v iskalnem dvojiskem drevesu poisce najmanjsi element
    public static int najmanjsi(Drevo drevo) throws Exception{
        //vrne najmanji element v iskalnem dvojiskem drevesu
        if (drevo.prazno())
          throw new Exception(" Drevo je prazno"); // v drevesu ni najmanjsega
       if (drevo.leviSin().prazno()) {
         return drevo.vrni(); // ce levega poddrevesa ni,je koren najmanjsi
       }
       return najmanjsi(drevo.leviSin()); //najmanjsi je v levem poddrevesu

   }

   // nerekurzivna metoda za iskanje najmanjsega elementa
   public static int min(Drevo drevo)throws Exception{
	   if(drevo.prazno()){
		   throw new Exception(" Drevo je prazno"); // v drevesu ni najmanjsega
	   }
	   else{
		   int min = drevo.vrni();
		   while(!drevo.prazno()){
			   min = drevo.vrni();
		       drevo = drevo.leviSin();
	       }
	       return min;
	   }
	}

   // metoda, ki v iskalnem dvojiskem drevesu poisce najvecji element
   public static int najvecji(Drevo drevo) throws Exception {
        // vrne najvecji element v iskalnem dvojiskem drevesu
        if (drevo.prazno())
         throw new Exception(" Drevo je prazno"); // v drevesu ni najvecjega
        if(drevo.desniSin().prazno()) {
         return drevo.vrni(); // ce desnega poddrevesa ni je koren najvecji
        }
       return najvecji(drevo.desniSin()); // najvecji je v desnem poddrevesu
   }

   // nerekurzivna metoda za iskanje najvecjega elementa
   public static int max(Drevo drevo)throws Exception{
	   if(drevo.prazno()){
		   throw new Exception(" Drevo je prazno"); // v praznem drevesu ga ni
	   }
	   else{
		  int max = drevo.vrni();
	      while(!drevo.prazno()){
		     max = drevo.vrni();
		     drevo = drevo.desniSin();
	      }
	      return max;
	  }
   }


///////////////////////////// VSTAVI //////////////////////////////////////

   // metoda vstavi, ki v iskalno dvojisko drevo vstavi nov element
   public  void vstaviVDrevo(int nPodatek){
       koren = vstaviVDrevo(koren, nPodatek);

   }

   public  Vozel vstaviVDrevo(Vozel koren, int nPodatek){
        if (koren == null){
            koren = new Vozel(nPodatek);
        }
        else{
            if (nPodatek < koren.vrniPodatek()){
               // nov vozel v levo poddrevo
               koren.nastaviLevi(vstaviVDrevo(koren.vrniLevi(), nPodatek));
               //koren.nastaviOce(vstaviVDrevo(koren.vrniLevi(), nPodatek));
            }
            else{
               // nov vozel v desno poddrevo
               koren.nastaviDesni(vstaviVDrevo(koren.vrniDesni(),nPodatek));
              // koren.nastaviOce(vstaviVDrevo(koren.vrniDesni(),nPodatek));
            }
        }

        return koren;
    }

///////////////////////////// BRISI ////////////////////////////////////////
    //metoda odstrani,ki iz iskalnega dvojiskega drevesa odstrani  element
    public void brisi(int nPodatek)throws Exception{
        koren = brisi(koren, nPodatek);
    }
    public Vozel brisi(Vozel koren, int nPodatek)throws Exception{
       if (koren == null){
          throw new Exception(" Drevo je prazno"); //ce je drevo prazno ne naredimo nic
       }
       else{
			if (nPodatek == koren.vrniPodatek()){ // nasli smo podatek
			return izbrisiVozel(koren);   // in ga zbrisemo
       }
       else if(nPodatek < koren.vrniPodatek()){ //podatek manjsi od korena
               // gremo rekurzivno v levo in primerjamo podatek
               koren.nastaviLevi(brisi(koren.vrniLevi(), nPodatek));
           }
           else{
			  // drugace gremo rekurzivno  v desno primerjamo podatek
              koren.nastaviDesni(brisi(koren.vrniDesni(), nPodatek));
           }
        }
        //vrnemo koren drevesa
        return koren;
    }

    // metoda za brisanje vozla  iz drevesa
    private static Vozel izbrisiVozel(Vozel koren) {
      if(koren.vrniLevi() == null){ // levega poddrevesa ni
      koren = koren.vrniDesni(); // nastavimo koren desnega poddrevesa
      }
      else if(koren.vrniDesni() == null) { //desnega poddrevesa ni
      koren = koren.vrniLevi(); // nastavimo koren levega poddrevesa
      }
      else{//imamo desno in levo poddrevo
         // poiscemo skrajno levi vozel v desnem poddrevesu,
         Vozel zadnji = poisciSkrajniLevi(koren.vrniDesni());
         //zamenjamo podatke
         koren.nastaviPodatek(zadnji.vrniPodatek());
         //zbrisemo skrajno levega
         koren.nastaviDesni(zbrisiSkrajniLevi(koren.vrniDesni()));
      }
      return koren;
    }

    //metoda,ki vrne skrajno levi vozel
    private static Vozel poisciSkrajniLevi(Vozel koren) {
       if(koren.vrniLevi() != null) { //levo poddrevo ni prazno
         //vrne skrajno levega
         return poisciSkrajniLevi(koren.vrniLevi());
       }
       else {
		 //najmanjsi je kar koren
         return koren;
       }
    }
    // izbrisemo skrajni levi vozel
    private static Vozel zbrisiSkrajniLevi(Vozel koren) {
       if(koren.vrniLevi() != null) { //levo poddrevo ni prazno
       	  //gremo v levo in ga zbrisemo
          koren.nastaviLevi(zbrisiSkrajniLevi(koren.vrniLevi()));
          return koren;
       }
       else{
		  //vrne koren desnega poddrevesa
          return koren.vrniDesni();
       }
   }
}

